@@ -106,13 +106,13 @@
 
 #define HASH_ID_KEY(n)		n->loc_id
 #define HASH_ID_NEXT(n)		n->next_id
-#define HASH_ID_EQ(a,b)		(a == b)
-#define HASH_ID_FN(k)		(k)
+#define HASH_ID_EQ(a,b)		a == b
+#define HASH_ID_FN(k)		k
 
 #define HASH_IP_KEY(n)		n->addr
 #define HASH_IP_NEXT(n)		n->next_ip
 #define HASH_IP_EQ(a,b)		ipa_equal(a,b)
-#define HASH_IP_FN(k)		ipa_hash(k)
+#define HASH_IP_FN(k)		ipa_hash32(k)
 
 static list bfd_proto_list;
 static list bfd_wait_list;
@@ -477,8 +477,15 @@ bfd_remove_session(struct bfd_proto *p, struct bfd_session *s)
 {
   ip_addr ip = s->addr;
 
+  /* Caller should ensure that request list is empty */
+
   birdloop_enter(p->loop);
 
+  /* Remove session from notify list if scheduled for notification */
+  /* No need for bfd_lock_sessions(), we are already protected by birdloop_enter() */
+  if (NODE_VALID(&s->n))
+    rem_node(&s->n);
+
   bfd_free_iface(s->ifa);
 
   rfree(s->tx_timer);
@@ -740,7 +747,7 @@ bfd_neigh_notify(struct neighbor *nb)
 
   if ((nb->scope > 0) && !n->req)
   {
-    ip_addr local = ipa_nonzero(n->local) ? n->local : nb->iface->addr->ip;
+    ip_addr local = ipa_nonzero(n->local) ? n->local : nb->ifa->ip;
     n->req = bfd_request_session(p->p.pool, n->addr, local, nb->iface, NULL, NULL);
   }
 
