@@ -13,11 +13,13 @@
 void ospf_dump_lsahdr(struct proto *p, struct ospf_lsa_header *lsa_n);
 void ospf_dump_common(struct proto *p, struct ospf_packet *op);
 void ospf_lsupd_send_list(struct ospf_neighbor *n, list * l);
-void ospf_lsupd_receive(struct ospf_lsupd_packet *ps,
+void ospf_lsupd_receive(struct ospf_packet *ps_i,
 			struct ospf_iface *ifa, struct ospf_neighbor *n);
-int ospf_lsupd_flood(struct ospf_neighbor *n, struct ospf_lsa_header *hn,
-		     struct ospf_lsa_header *hh, struct ospf_iface *iff,
-		     struct ospf_area *oa, int rtl);
-void ospf_lsupd_flush_nlsa(struct top_hash_entry *en, struct ospf_area *oa);
+int ospf_lsupd_flood(struct proto_ospf *po,
+		     struct ospf_neighbor *n, struct ospf_lsa_header *hn,
+		     struct ospf_lsa_header *hh, u32 domain, int rtl);
+void ospf_lsupd_flush_nlsa(struct proto_ospf *po, struct top_hash_entry *en);
+int ospf_lsa_flooding_allowed(struct ospf_lsa_header *lsa, u32 domain, struct ospf_iface *ifa);
+
 
 #endif /* _BIRD_OSPF_LSUPD_H_ */
